#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2020, Florian Schanda                         ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

from miss_hit.graph import Graph, Vertex_Root
from miss_hit_core.m_ast import *
from miss_hit_core.errors import Message_Handler, ICE


class Vertex(Vertex_Root):
    def __init__(self, graph, n_node):
        assert isinstance(n_node, Node)

        super().__init__(graph)

        self.n_node = n_node

    def dot_label(self):
        return "%s" % self.n_node.__class__.__name__


class CFG_Context:
    def __init__(self, v_entry=None):
        assert v_entry is None or isinstance(v_entry, Vertex_Root)

        self.v_entry = v_entry
        self.l_exits = []
        self.l_loop_breaks = []
        self.l_loop_continues = []

    def merge_loops(self, other):
        assert isinstance(other, CFG_Context)

        self.l_loop_breaks += other.l_loop_breaks
        self.l_loop_continues += other.l_loop_continues

    def merge_exits(self, other):
        assert isinstance(other, CFG_Context)
        self.l_exits += other.l_exits


def build_cfg_statement(graph, n_statement):
    assert isinstance(graph, Graph)
    assert isinstance(n_statement, (Statement,
                                    Metric_Justification_Pragma))

    ctx = CFG_Context(Vertex(graph, n_statement))

    if isinstance(n_statement, (Compound_Assignment_Statement,
                                Global_Statement,
                                Import_Statement,
                                Naked_Expression_Statement,
                                Persistent_Statement,
                                Simple_Assignment_Statement,
                                Metric_Justification_Pragma)):
        # All of these are simple statements. One entry, one obvious
        # exit.
        ctx.l_exits.append(ctx.v_entry)

    elif isinstance(n_statement, If_Statement):
        # If statements chain together the actions.
        current_link = ctx.v_entry
        for n_action in n_statement.l_actions:
            v_action = Vertex(graph, n_action)
            graph.add_edge(current_link, v_action)
            current_link = v_action

            action_ctx = build_cfg_sos(graph, n_action.n_body)
            ctx.merge_loops(action_ctx)
            ctx.merge_exits(action_ctx)
            if action_ctx.v_entry:
                graph.add_edge(v_action, action_ctx.v_entry)

        if not n_statement.has_else:
            # Add an implicit path for else, if not present
            ctx.l_exits.append(current_link)

    elif isinstance(n_statement, Switch_Statement):
        # Case statements in MATLAB are chained, because there is no
        # semantic check that the options do not overlap; hence they
        # must be evaluated in sequence. This is the same as above. In
        # the future, for the MISS_HIT subset, we could make sure
        # there is no overlap.

        current_link = ctx.v_entry
        for n_action in n_statement.l_actions:
            v_action = Vertex(graph, n_action)
            graph.add_edge(current_link, v_action)
            current_link = v_action

            action_ctx = build_cfg_sos(graph, n_action.n_body)
            ctx.merge_loops(action_ctx)
            ctx.merge_exits(action_ctx)
            if action_ctx.v_entry:
                graph.add_edge(v_action, action_ctx.v_entry)

        if not n_statement.has_otherwise:
            # Add an implicit path for otherwise, if not present
            ctx.l_exits.append(current_link)

    elif isinstance(n_statement, (For_Loop_Statement,
                                  While_Statement)):
        # Loops add a loop back to the loop statement. There are two
        # paths out of the loop statement: one into the loop and one
        # to the next statement.
        #
        # Any break statements found in the body are processed here.
        body_ctx = build_cfg_sos(graph, n_statement.n_body)

        if body_ctx.v_entry:
            graph.add_edge(ctx.v_entry, body_ctx.v_entry)

        for src in body_ctx.l_exits:
            graph.add_edge(src, ctx.v_entry)
        for src in body_ctx.l_loop_continues:
            graph.add_edge(src, ctx.v_entry)

        ctx.l_exits = [ctx.v_entry] + ctx.l_loop_breaks

    elif isinstance(n_statement, Break_Statement):
        ctx.l_loop_breaks.append(ctx.v_entry)

    elif isinstance(n_statement, Continue_Statement):
        ctx.l_loop_continues.append(ctx.v_entry)

    elif isinstance(n_statement, Return_Statement):
        graph.add_edge(ctx.v_entry, graph.get_named_vertex("end"))

    elif isinstance(n_statement, SPMD_Statement):
        spmd_ctx = build_cfg_sos(graph, n_statement.n_body)
        ctx.merge_loops(spmd_ctx)
        ctx.merge_exits(spmd_ctx)

        if spmd_ctx.v_entry:
            graph.add_edge(ctx.v_entry, spmd_ctx.v_entry)

    elif isinstance(n_statement, Try_Statement):
        try_ctx = build_cfg_sos(graph, n_statement.n_body)
        ctx.merge_loops(try_ctx)
        ctx.merge_exits(try_ctx)

        if n_statement.n_handler:
            handler_ctx = build_cfg_sos(graph, n_statement.n_handler)
            ctx.merge_loops(handler_ctx)
            ctx.merge_exits(handler_ctx)
            if handler_ctx.v_entry:
                graph.add_edge(ctx.v_entry, handler_ctx.v_entry)
        else:
            ctx.l_exits.append(ctx.v_entry)

        if try_ctx.v_entry:
            graph.add_edge(ctx.v_entry, try_ctx.v_entry)

    else:
        raise ICE("unknown statement kind %s" %
                  n_statement.__class__.__name__)

    return ctx


def build_cfg_sos(graph, n_sos):
    assert isinstance(graph, Graph)
    assert isinstance(n_sos, Sequence_Of_Statements)

    ctx = CFG_Context()

    for n_statement in n_sos.l_statements:
        statement_ctx = build_cfg_statement(graph, n_statement)

        # Set entry point to the first entry point in the statement
        # list.
        if ctx.v_entry is None:
            ctx.v_entry = statement_ctx.v_entry

        # Link the previous exits to the just processed statement. We
        # then update our exists to the exits of the processed
        # statement.
        for src in ctx.l_exits:
            graph.add_edge(src, statement_ctx.v_entry)
        ctx.l_exits = statement_ctx.l_exits

        # Accumulate loop breaks
        ctx.merge_loops(statement_ctx)

    return ctx


def build_cfg(n_fdef):
    assert isinstance(n_fdef, (Function_Definition,
                               Script_File))

    graph = Graph()
    v_start = Vertex_Root(graph, "start")
    v_end = Vertex_Root(graph, "end")

    if isinstance(n_fdef, Function_Definition):
        ctx = build_cfg_sos(graph, n_fdef.n_body)
    else:
        ctx = build_cfg_sos(graph, n_fdef.n_statements)

    assert len(ctx.l_loop_breaks) == 0

    if ctx.v_entry:
        graph.add_edge(v_start, ctx.v_entry)
        for v_exit in ctx.l_exits:
            graph.add_edge(v_exit, v_end)
    else:
        graph.add_edge(v_start, v_end)

    return graph


def debug_cfg(cunit, mh):
    assert isinstance(cunit, Compilation_Unit)
    assert isinstance(mh, Message_Handler)

    class Function_Visitor(AST_Visitor):
        def visit(self, node, n_parent, relation):
            if isinstance(node, (Function_Definition,
                                 Script_File)):
                # pylint: disable=unused-variable
                cfg = build_cfg(node)
                # if isinstance(n_fdef, Function_Definition):
                #     graph.debug_write_dot(str(n_fdef.n_sig.n_name))
                # else:
                #     graph.debug_write_dot(n_fdef.name)

    cunit.visit(None, Function_Visitor(), "Root")



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2020-2021, Florian Schanda                    ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

from miss_hit_core.m_entity_root import Entity
from miss_hit_core.m_ast import *


##############################################################################
# Symbol table
##############################################################################

class Scope:
    # https://www.mathworks.com/help/matlab/matlab_prog/function-precedence-order.html
    #
    # Note that names (for functions) are resolved in this order:
    #
    #  1. Variable in the current workspace
    #  2. Imported name
    #  3. Nested functions in the current function
    #  4. Local function in the current file
    #  5. Wildcard imported name
    #  6. Private function (functions in a directory 'private' relative
    #     to the current file)
    #  7. Object functions (todo: unclear. are these methods?)
    #  8. Class constructors in @ directories (i.e. @foo/foo.m)
    #  9. Loaded simulink models (todo: unclear what this means)
    # 10. Functions in the current directory
    # 11. Functions on path, based on the path ordering
    #
    # Since calling Simulink models in a context that is not the
    # MATLAB command prompt, we can ignore (9).
    #
    # The 'current directory' (10) will be somwehat irrelevant as
    # well, since a MISS_HIT entry-point should produce a sensible
    # path.
    #
    # If there is a conflict within the same directory, it is resolved
    # using this order of preference:
    #
    # 1. Built-in function
    # 2. MEX function
    # 3. Simulink models that are not loaded (preferring slx over mdl)
    # 4. Stateflow chart with a .sfx extension
    # 5. App file (.mlapp) created by the App Designer
    # 6. mlx files
    # 7. p files
    # 8. m files
    #
    # Since the context of MISS_HIT is industrial, and generally
    # assuming this are sane; any confliuct here will just trigger an
    # error.
    #
    # Further, only (1) and (8) will be considered, since those are
    # the only sane things we can analyze.
    #
    # Also see:
    # https://www.mathworks.com/help/matlab/matlab_oop/scoping-classes-with-packages.html
    def __init__(self):
        self.names = [{}]

    def dump(self, filename=None):
        title = "Symbol table"
        if filename:
            title += " for %s" % filename
        title += " with %u active scopes:" % len(self.names)
        print(title)
        for level, names in enumerate(self.names, 1):
            print("=== Level %u ===" % level)
            for name in sorted(names):
                e_sym = names[name]
                e_sym.dump()

    def push(self):
        self.names.append({})

    def pop(self):
        if self.names:
            self.names.pop()
        else:
            raise ICE("tried to pop empty scope stack")

    def register(self, mh, entity):
        assert isinstance(mh, Message_Handler)
        assert isinstance(entity, Entity)

        if entity.name in self.names[-1]:
            # TODO: improve error message
            mh.error(entity.loc(),
                     "duplicate definition")

        else:
            self.names[-1][entity.name] = entity

    def lookup_str(self, name):
        assert isinstance(name, str)
        for layer in reversed(self.names):
            if name in layer:
                return layer[name]
        return None

    def lookup_token(self, mh, token, fatal=True):
        assert isinstance(mh, Message_Handler)
        assert isinstance(token, MATLAB_Token)
        assert token.kind in ("IDENTIFIER", "KEYWORD")
        assert isinstance(fatal, bool)

        entity = self.lookup_str(token.value)

        if fatal and entity is None:
            mh.error(token.location,
                     "unknown name")

        return entity

    def lookup_identifier(self, mh, n_ident, fatal=True):
        assert isinstance(mh, Message_Handler)
        assert isinstance(n_ident, Identifier)
        assert isinstance(fatal, bool)

        return self.lookup_token(n_ident.t_ident, fatal)

    def import_visible_names(self, other):
        assert isinstance(other, Scope)

        # Very important assumption here is that any duplicate
        # definitions have already been checked.

        for e_sym in other.names[-1].values():
            if not e_sym.is_externally_visible:
                continue
            if e_sym.name in self.names[-1]:
                raise ICE("attempted to import already already known "
                          "symbol %s" % e_sym.name)
            self.names[-1][e_sym.name] = e_sym


class Package_Entity(Entity):
    # Packages are directories starting with a '+'. There are some
    # fine points here: firstly, can split this over multiple
    # directories (i.e. if a and b are on your path, and both contain
    # +foo, then this is not an error). Secondly, any directory can
    # also contain a private directory: the contents are available in
    # that parent directory only, and do not extend over the entire
    # package.
    #
    # Hence our packages (which are entities) are actually collections
    # of package directories (which are not entities).
    def __init__(self, name):
        super().__init__(name               = name,
                         externally_visible = True)

        self.directories = []
        self.class_directories = []
        self.children = {}

    def add_directory(self, pkgdir):
        assert isinstance(pkgdir, Package_Directory)
        self.directories.append(pkgdir)

    def add_class_directory(self, clsdir):
        assert isinstance(clsdir, Class_Directory)
        self.class_directories.append(clsdir)

    def add_child_package(self, pkg):
        assert isinstance(pkg, Package_Entity)

        if pkg.name in self.children:
            raise ICE("NIY: merging packages")

        else:
            self.children[pkg.name] = pkg


class Directory:
    def __init__(self, dirname):
        assert isinstance(dirname, str)
        self.dirname = dirname

        self.private_directory = None

    def set_private_directory(self, prvdir):
        assert isinstance(prvdir, Private_Directory)
        if self.private_directory is not None:
            raise ICE("set duplicate private directory")
        self.private_directory = prvdir

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           self.dirname)


class Package_Directory(Directory):
    pass


class Global_Scope(Package_Directory):
    pass


class Class_Directory(Directory):
    pass


class Private_Directory(Directory):
    def set_private_directory(self, prvdir):
        raise ICE("set private directory on private directory")


##############################################################################
# Entities
##############################################################################

class Function_Entity(Entity):
    def __init__(self, n_fsig):
        assert isinstance(n_fsig, Function_Signature)
        super().__init__(name = str(n_fsig.n_name))

        self.n_definition = None
        # The actual function body.

        self.n_declaration = n_fsig
        # The original declaration. While usually n_definition.n_sig
        # is n_declaration, this is not true for separates. In this
        # case the decclaration refers to the signature in the main
        # class file, and the definition refers to the separate body.


class Class_Entity(Entity):
    def __init__(self, n_classdef):
        assert isinstance(n_classdef, Class_Definition)
        super().__init__(name               = str(n_classdef.n_name),
                         externally_visible = True)
        # Note: it is not legal to have a local or private class.

        self.n_definition = n_classdef
        self.n_definition.entity = self
        # The actual definition

        self.el_super = []
        # Entity list of super-classes

    def dump(self):
        print("Class Entity (%s)" % self.name)
        print("  definition: %s" % self.n_definition.loc())



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2020-2021, Florian Schanda                    ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

import os

from miss_hit_core import command_line
from miss_hit_core import work_package
from miss_hit_core import cfg_tree
from miss_hit_core.m_ast import *
from miss_hit_core.errors import (Error,
                                  Message_Handler,
                                  HTML_Message_Handler,
                                  JSON_Message_Handler)
from miss_hit_core.m_lexer import MATLAB_Lexer
from miss_hit_core.m_parser import MATLAB_Parser
from miss_hit_core.m_language_builtins import BUILTIN_FUNCTIONS

from miss_hit.m_sem import sem_pass_1


class Stage_1_Linting(AST_Visitor):
    """ Checks which do not require semantic analysis """
    def __init__(self, mh):
        self.mh = mh

    def visit(self, node, n_parent, relation):
        if isinstance(node, Compilation_Unit):
            self.check_compilation_unit(node)

    def check_compilation_unit(self, n_cu):
        assert isinstance(n_cu, Compilation_Unit)

        # Check that filename matches the primary entity
        if isinstance(n_cu, Function_File):
            self.check_filename(file_name = n_cu.name,
                                kind      = "function",
                                ent_name  = n_cu.l_functions[0].n_sig.n_name)
        elif isinstance(n_cu, Class_File):
            self.check_filename(file_name = n_cu.name,
                                kind      = "class",
                                ent_name  = n_cu.n_classdef.n_name)

        # Check that Contents.m doesn't contain anything but
        # comments. We do this case-insensitive to make sure this
        # nobody slips in a "ConTENTS.m" file on Linux that would
        # cause issues in Windows.
        if n_cu.name.lower() == "contents.m":
            error_loc = None
            if isinstance(n_cu, Script_File):
                # No need to check functions, a script file with no
                # statements would be a functin file OR a blank file
                is_blank = len(n_cu.n_statements.l_statements) == 0
                if not is_blank:
                    error_loc = n_cu.n_statements.l_statements[0].loc()

            elif isinstance(n_cu, Function_File):
                # Function files are always a problem
                is_blank = False
                error_loc = n_cu.l_functions[0].loc()

            elif isinstance(n_cu, Class_File):
                # Class files are also always a problem
                is_blank = False
                error_loc = n_cu.n_classdef.loc()

            if not is_blank:
                self.mh.check(error_loc,
                              "a Contents.m file must only contain comments",
                              "low")

    def check_filename(self, file_name, kind, ent_name):
        assert isinstance(file_name, str)
        assert isinstance(ent_name, Name) and ent_name.is_simple_dotted_name()

        base_filename, ext = os.path.splitext(file_name)

        if ext != ".m":
            # This check is not applicable for MATLAB embedded in
            # Simulink models
            pass
        elif base_filename != str(ent_name):
            self.mh.check(ent_name.loc(),
                          "%s name does not match the filename %s" %
                          (kind, base_filename),
                          "low")


class MH_Lint_Result(work_package.Result):
    def __init__(self, wp, sem=None):
        super().__init__(wp, True)
        self.sem = sem


class MH_Lint(command_line.MISS_HIT_Back_End):
    def __init__(self, options):
        super().__init__("MH Lint")
        self.perform_sem = options.entry_point is not None
        self.debug_show_st = options.debug_show_global_symbol_table

    @classmethod
    def process_wp(cls, wp):
        # Create lexer
        lexer = MATLAB_Lexer(wp.mh,
                             wp.get_content(),
                             wp.filename,
                             wp.blockname)
        if wp.cfg.octave:
            lexer.set_octave_mode()
        if not wp.cfg.pragmas:
            lexer.process_pragmas = False
        if len(lexer.text.strip()) == 0:
            return MH_Lint_Result(wp)

        # Create parse tree
        try:
            parser = MATLAB_Parser(wp.mh, lexer, wp.cfg)
            n_cu = parser.parse_file()
        except Error:
            return MH_Lint_Result(wp)

        # Check compilation units for shadowing a built-in
        base_name = os.path.splitext(n_cu.name)[0]
        dir_name = os.path.basename(os.path.dirname(
            os.path.abspath(wp.filename)))
        if wp.cfg.active("builtin_shadow") and \
           base_name in BUILTIN_FUNCTIONS and \
           not (dir_name.startswith("+") or dir_name.startswith("@")):
            wp.mh.check(n_cu.loc(),
                        "this file shadows built-in %s which is very naughty"
                        % base_name,
                        "high")

        # Initial checks
        n_cu.visit(None, Stage_1_Linting(wp.mh), "Root")

        # First pass of semantic analysis
        entrypoint = cfg_tree.get_entry_point(wp.options.entry_point)
        sem = sem_pass_1(wp.mh, entrypoint, n_cu)

        return MH_Lint_Result(wp, sem)

    def process_result(self, result):
        if not self.perform_sem:
            return
        if not isinstance(result, MH_Lint_Result):
            return
        if result.sem is None:
            return

        if self.debug_show_st:
            result.sem.scope.dump(result.wp.filename)

    def post_process(self):
        pass


def main_handler():
    clp = command_line.create_basic_clp()

    # Extra output options
    clp["output_options"].add_argument(
        "--html",
        default=None,
        help="Write report to given file as HTML")
    clp["output_options"].add_argument(
        "--json",
        default=None,
        help="Produce JSON report")

    # Extra debug options
    clp["debug_options"].add_argument(
        "--debug-show-global-symbol-table",
        default=False,
        action="store_true",
        help="Show global symbol table")

    options = command_line.parse_args(clp)

    if options.html:
        if options.json:
            clp["ap"].error("Cannot produce JSON and HTML at the same time")
        if os.path.exists(options.html) and not os.path.isfile(options.html):
            clp["ap"].error("Cannot write to %s: it is not a file" %
                            options.html)
        mh = HTML_Message_Handler("lint", options.html)
    elif options.json:
        if os.path.exists(options.json) and not os.path.isfile(options.json):
            clp["ap"].error("Cannot write to %s: it is not a file" %
                            options.json)
        mh = JSON_Message_Handler("lint", options.json)
    else:
        mh = Message_Handler("lint")

    mh.show_context = not options.brief
    mh.show_style   = False
    mh.show_checks  = True
    mh.autofix      = False

    lint_backend = MH_Lint(options)
    command_line.execute(mh, options, {}, lint_backend)


def main():
    command_line.ice_handler(main_handler)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2020, Florian Schanda                         ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

# Simple helper for performing type inference. You start with all
# options, and then you can exclude specific classes trees or assert
# the option must be in a specific class tree.

import inspect
from copy import copy

from miss_hit_core import m_types
from miss_hit import graph


def build_masks():
    cls_graph = graph.Graph()
    nv_map = {}

    rv = {
        "leafs" : set(),
        "mask"  : {}
    }

    for name, c in inspect.getmembers(m_types, inspect.isclass):
        if not issubclass(c, m_types.Type):  # pragma: no cover
            continue
        nv_map[name] = graph.Vertex_Root(cls_graph, name)
    for name, c in inspect.getmembers(m_types, inspect.isclass):
        for base in c.__bases__:
            if issubclass(base, m_types.Type):
                cls_graph.add_edge(nv_map[base.__name__],
                                   nv_map[name])

    # Get a set of leaf nodes (since these are the only useful results
    # of inference)
    for vrt in cls_graph.vertices:
        if not vrt.out_edges:
            rv["leafs"].add(vrt.name)

    # Get mask for each type. Not ideal, ideally we have a transitive
    # closure algorithm in the graph class, but this works for now.
    def add_mask(base, vrt):
        rv["mask"][base].add(vrt.name)
        for child in vrt.out_edges:
            add_mask(base, child)
    for vrt in cls_graph.vertices:
        rv["mask"][vrt.name] = set()
        add_mask(vrt.name, vrt)

        # Reduce mask by just including the leaves
        rv["mask"][vrt.name] &= rv["leafs"]

    return rv


class Type_Inference:
    MASKS = build_masks()

    def __init__(self):
        self.options = copy(Type_Inference.MASKS["leafs"])

    def is_resolved(self):
        return len(self.options) == 1

    def is_conflicted(self):
        return len(self.options) == 0

    def assert_positive(self, choice):
        assert issubclass(choice, m_types.Type)
        self.options &= Type_Inference.MASKS["mask"][choice.__name__]

    def assert_negative(self, choice):
        assert issubclass(choice, m_types.Type)
        self.options -= Type_Inference.MASKS["mask"][choice.__name__]

    def dump(self):
        print("Type inference")
        print("  resolved: %s" % self.is_resolved())
        print("  conflict: %s" % self.is_conflicted())
        print("  options:  %s" % ", ".join(sorted(self.options)))


def sanity_test():
    inf = Type_Inference()

    inf.dump()

    inf.assert_positive(m_types.Numeric_Type)
    inf.dump()

    inf.assert_negative(m_types.Integer_Type)
    inf.dump()

    inf.assert_positive(m_types.Scalar_Type)
    inf.dump()


if __name__ == "__main__":
    sanity_test()



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2021, Florian Schanda                         ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

import functools
import operator
import json

from miss_hit_core import command_line
from miss_hit_core import work_package
from miss_hit_core.m_ast import *
from miss_hit_core.errors import (Error,
                                  Message_Handler)
from miss_hit_core.m_lexer import MATLAB_Lexer
from miss_hit_core.m_parser import MATLAB_Parser


class MH_Trace_Result(work_package.Result):
    def __init__(self, wp, tracing=None):
        super().__init__(wp, True)
        self.tracing = tracing


class Function_Visitor(AST_Visitor):
    def __init__(self, mh, cu):
        assert isinstance(mh, Message_Handler)
        assert isinstance(cu, Compilation_Unit)

        self.tag_stack = []
        self.mh = mh
        self.tracing = {}
        self.name_prefix = cu.get_name_prefix()

    def get_test_tags(self, node):
        n_tags = node.get_attribute("TestTags")
        rv = set()

        if n_tags is None:
            return rv

        if not isinstance(n_tags.n_value, (Cell_Expression,
                                           Matrix_Expression)):
            self.mh.error(n_tags.n_value.loc(),
                          "TestTags value must be a cell or matrix",
                          fatal=False)
            return rv

        n_row_list = n_tags.n_value.n_content
        if len(n_row_list.l_items) != 1:
            self.mh.error(n_tags.n_value.loc(),
                          "TestTags value must only contain precisely"
                          " 1 row",
                          fatal=False)
            return rv

        for n_tag in n_row_list.l_items[0].l_items:
            if isinstance(n_tag, (Char_Array_Literal,
                                  String_Literal)):
                rv.add(n_tag.t_string.value)
            else:
                self.mh.error(n_tag.loc(),
                              "TestTags value must be a string or"
                              " carray literal",
                              fatal=False)

        return rv

    def visit(self, node, n_parent, relation):
        # Deal with the tag stack first
        if isinstance(node, Function_Definition):
            self.tag_stack.append(set())
        elif isinstance(node, Class_Definition):
            self.tag_stack.append(self.get_test_tags(node))
        elif isinstance(node, Compilation_Unit):
            self.tag_stack.append(set())
        elif isinstance(node, Special_Block) and node.kind() == "methods":
            self.tag_stack.append(self.get_test_tags(node))

        # Amend current tag stack if we get a tag pragma
        if isinstance(node, Tag_Pragma):
            self.tag_stack[-1] |= node.get_tags()

    def visit_end(self, node, n_parent, relation):
        # Create entry for tracing
        if isinstance(node, Function_Definition):
            name = self.name_prefix + node.get_local_name()
            self.tracing[name] = {
                "source": node.loc().to_json(detailed=False),
                "tags"  : sorted(functools.reduce(operator.or_,
                                                  self.tag_stack,
                                                  set()))
            }

        if isinstance(node, (Definition,
                             Compilation_Unit)):
            self.tag_stack.pop()
        elif isinstance(node, Special_Block) and node.kind() == "methods":
            self.tag_stack.pop()


class MH_Trace(command_line.MISS_HIT_Back_End):
    def __init__(self, options):
        super().__init__("MH Trace")
        self.tracing = {}
        self.options = options

    @classmethod
    def process_wp(cls, wp):
        # Create lexer
        lexer = MATLAB_Lexer(wp.mh,
                             wp.get_content(),
                             wp.filename,
                             wp.blockname)
        if wp.cfg.octave:
            lexer.set_octave_mode()
        if not wp.cfg.pragmas:
            lexer.process_pragmas = False
        if len(lexer.text.strip()) == 0:
            return MH_Trace_Result(wp)

        # Create parse tree
        try:
            parser = MATLAB_Parser(wp.mh, lexer, wp.cfg)
            n_cu = parser.parse_file()
        except Error:
            return MH_Trace_Result(wp)

        visitor = Function_Visitor(wp.mh, n_cu)
        n_cu.visit(None, visitor, "Root")

        # Return results
        return MH_Trace_Result(wp, visitor.tracing)

    def process_result(self, result):
        if result.tracing:
            self.tracing.update(result.tracing)

    def post_process(self):
        if self.options.by_tag:
            out = {}
            for unitname in sorted(self.tracing):
                source = self.tracing[unitname]["source"]
                tags   = self.tracing[unitname]["tags"]
                for tag in tags:
                    if tag not in out:
                        out[tag] = []
                    out[tag].append({"name"   : unitname,
                                     "source" : source})
        else:
            out = self.tracing

        with open(self.options.json, "w") as fd:
            json.dump(out, fd, indent=4, sort_keys=True)


def main_handler():
    clp = command_line.create_basic_clp()

    clp["output_options"].add_argument(
        "--json",
        default="mh_trace.json",
        help="name of the JSON report (by default mh_trace.json)")
    clp["output_options"].add_argument(
        "--by-tag",
        action="store_true",
        default=False,
        help=("group tracing information by tag; by default we group by"
              " unit name first"))

    options = command_line.parse_args(clp)

    mh = Message_Handler("trace")

    mh.show_context = not options.brief
    mh.show_style   = False
    mh.show_checks  = False
    mh.autofix      = False

    trace_backend = MH_Trace(options)
    command_line.execute(mh, options, {}, trace_backend,
                         process_tests=True)


def main():
    command_line.ice_handler(main_handler)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2020, Florian Schanda                         ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

import os
import json
import subprocess

from miss_hit_core import command_line
from miss_hit_core import work_package
from miss_hit_core import m_ast
from miss_hit_core.errors import Message_Handler, Error, Location, ICE
from miss_hit_core.m_lexer import MATLAB_Lexer
from miss_hit_core.m_parser import MATLAB_Parser

from miss_hit import goto_ast


def set_location(sym, loc):
    assert isinstance(sym, goto_ast.Irep)
    assert isinstance(loc, Location)

    sloc = goto_ast.Irep("")
    sloc.set_attribute("file", loc.filename)
    if loc.line is not None:
        sloc.set_attribute("line", str(loc.line))
    if loc.col_start is not None:
        sloc.set_attribute("column", "%u:%u" % (loc.col_start,
                                                loc.col_end))

    sym.named_sub["#source_location"] = sloc


def make_type():
    # One of the key limitations of the MVP: we assume that everything
    # is a 32-bit signed integer.
    return goto_ast.SignedBV_Type(32)


def compile_name(mh, gst, stab, n_name):
    assert isinstance(mh, Message_Handler)
    assert isinstance(gst, goto_ast.GOTO_Symbol_Table)
    assert isinstance(stab, dict)
    assert isinstance(n_name, m_ast.Name)

    if isinstance(n_name, m_ast.Identifier):
        typ = make_type()
        pretty_name = str(n_name)
        mangled_name = stab[pretty_name]

        sym = goto_ast.Symbol_Expr(typ, mangled_name)
        set_location(sym, n_name.loc())
        return sym

    else:
        mh.error(n_name.loc(),
                 "mh_bmc does not %s names yet" %
                 n_name.__class__.__name__)


def compile_expression(mh, gst, stab, n_expr):
    assert isinstance(mh, Message_Handler)
    assert isinstance(gst, goto_ast.GOTO_Symbol_Table)
    assert isinstance(stab, dict)
    assert isinstance(n_expr, m_ast.Expression)

    if isinstance(n_expr, m_ast.Binary_Operation):
        typ = make_type()
        lhs = compile_expression(mh, gst, stab, n_expr.n_lhs)
        rhs = compile_expression(mh, gst, stab, n_expr.n_rhs)
        if n_expr.t_op.value == "+":
            sym = goto_ast.Plus_Expr(typ, [lhs, rhs])
            set_location(sym, n_expr.t_op.location)
            return sym
        else:
            mh.error(n_expr.t_op.location,
                     "mh_bmc does not the binary %s operation yet" %
                     n_expr.t_op.value)

    elif isinstance(n_expr, m_ast.Name):
        return compile_name(mh, gst, stab, n_expr)

    elif isinstance(n_expr, m_ast.Number_Literal):
        try:
            int_val = int(n_expr.t_value.value)
        except ValueError:
            mh.error(n_expr.t_value.location,
                     "mh_bmc only supports integer literals so far")
        typ = make_type()
        sym = goto_ast.Constant_Expr(typ, "%x" % int_val)
        set_location(sym, n_expr.t_value.location)
        return sym

    else:
        mh.error(n_expr.loc(),
                 "mh_bmc does not %s expressions yet" %
                 n_expr.__class__.__name__)


def compile_simple_assignment_statement(mh, gst, stab, n_stmt):
    assert isinstance(mh, Message_Handler)
    assert isinstance(gst, goto_ast.GOTO_Symbol_Table)
    assert isinstance(stab, dict)
    assert isinstance(n_stmt, m_ast.Simple_Assignment_Statement)

    typ = make_type()
    lhs = compile_name(mh, gst, stab, n_stmt.n_lhs)
    rhs = compile_expression(mh, gst, stab, n_stmt.n_rhs)

    sym_asn = goto_ast.Side_Effect_Expr_Assign(typ, lhs, rhs)
    sym_code = goto_ast.Code_Expression(sym_asn)
    return sym_code


def compile_statement(mh, gst, stab, n_stmt):
    assert isinstance(mh, Message_Handler)
    assert isinstance(gst, goto_ast.GOTO_Symbol_Table)
    assert isinstance(stab, dict)
    assert isinstance(n_stmt, m_ast.Statement)

    if isinstance(n_stmt, m_ast.Simple_Assignment_Statement):
        return compile_simple_assignment_statement(mh, gst, stab, n_stmt)

    else:
        mh.error(n_stmt.loc(),
                 "mh_bmc does not support %s yet" %
                 n_stmt.__class__.__name__)


def compile_sequence_of_statements(mh, gst, stab, n_seq):
    assert isinstance(mh, Message_Handler)
    assert isinstance(gst, goto_ast.GOTO_Symbol_Table)
    assert isinstance(stab, dict)
    assert isinstance(n_seq, m_ast.Sequence_Of_Statements)

    sym = goto_ast.Code_Block()

    for n_statement in n_seq.l_statements:
        sym.add_statement(compile_statement(mh, gst, stab, n_statement))

    return sym


def compile_function(mh, gst, n_fdef):
    assert isinstance(mh, Message_Handler)
    assert isinstance(gst, goto_ast.GOTO_Symbol_Table)
    assert isinstance(n_fdef, m_ast.Function_Definition)

    function_name = str(n_fdef.n_sig.n_name)

    sym_fn_type = goto_ast.Code_Type()
    stab = {}

    # Create names for the function inputs
    for n_param in n_fdef.n_sig.l_inputs:
        pretty_name = str(n_param)
        mangled_name = "::".join([function_name, "in", pretty_name])

        sym = goto_ast.Symbol(mangled_name, pretty_name)
        sym.is_parameter = True
        sym.value = goto_ast.Irep("nil")
        sym.typ = make_type()
        sym.location = n_param.loc()
        stab[pretty_name] = mangled_name
        gst.add_symbol(sym)

        sym_fn_type.add_parameter(goto_ast.Parameter(mangled_name, sym.typ))

    # Create names for the function outputs. Goto supports void
    # functions and functions with a single return. Right now we only
    # deal with non-void functions as that maps more or less 1:1.
    if len(n_fdef.n_sig.l_outputs) != 1:
        mh.error(n_fdef.loc(),
                 "mh_bmc supports only functions returning exactly one value")
    the_output = None
    for n_param in n_fdef.n_sig.l_outputs:
        pretty_name = str(n_param)
        mangled_name = "::".join([function_name, "out", pretty_name])

        sym = goto_ast.Symbol(mangled_name, pretty_name)
        sym.value = goto_ast.Irep("nil")
        sym.typ = make_type()
        sym.location = n_param.loc()
        stab[pretty_name] = mangled_name
        gst.add_symbol(sym)
        the_output = mangled_name

        sym_fn_type.set_return_type(sym.typ)

    # Complain about bits that we don't support yet
    for n_vld in n_fdef.l_validation:
        mh.error(n_vld.loc(), "not supported yet")
    for n_nst in n_fdef.l_nested:
        mh.error(n_nst.loc(), "nested function not supported yet")

    # Translate body
    sym = goto_ast.Symbol(function_name)
    sym.typ = sym_fn_type
    sym.value = compile_sequence_of_statements(mh, gst, stab, n_fdef.n_body)

    # Bolt on the function return
    sym_ret = goto_ast.Code_Return(goto_ast.Symbol_Expr(make_type(),
                                                        the_output))
    sym.value.add_statement(sym_ret)

    # Add to the symbol table
    gst.add_symbol(sym)


def compile_file(mh, n_tree):
    assert isinstance(mh, Message_Handler)
    assert isinstance(n_tree, m_ast.Node)

    if not isinstance(n_tree, m_ast.Function_File):
        mh.error(n_tree.loc(),
                 "mh_bmc only supports function files right now")

    gst = goto_ast.GOTO_Symbol_Table()

    # Compile all functions
    for n_fdef in n_tree.l_functions:
        compile_function(mh, gst, n_fdef)

    return gst


def analyze(mh, new_filename, lexer, n_file):
    assert isinstance(mh, Message_Handler)
    assert isinstance(new_filename, str)
    assert new_filename.endswith(".json_symtab")
    assert isinstance(lexer, MATLAB_Lexer)
    assert isinstance(n_file, m_ast.Function_File)

    def fmt_value(trace_value):
        assert trace_value["type"] == "int"
        return trace_value["data"]

    for n_fdef in n_file.l_functions:
        function_name = str(n_fdef.n_sig.n_name)

        cmd = ["cbmc",
               new_filename,
               "--json-ui",
               "--function", function_name,
               "--bounds-check",
               "--div-by-zero-check",
               "--signed-overflow-check",
               "--unsigned-overflow-check",
               "--float-overflow-check",
               "--nan-check"]
        cbmc_result = subprocess.run(cmd,
                                     check=False,
                                     capture_output=True,
                                     encoding="utf-8")
        assert cbmc_result.stderr == ""

        if cbmc_result.returncode == 0:
            mh.info(n_fdef.loc(), "verification successful")
            continue

        cbmc_msg = json.loads(cbmc_result.stdout)

        results = None
        for msg in cbmc_msg:
            if "result" in msg:
                if results is None:
                    results = msg["result"]
                else:
                    raise ICE("duplicate results")
        if results is None:
            raise ICE("result not found in cbmc output")

        for item in results:
            assert item["status"] == "FAILURE"

            fail_loc = item["trace"][-1]
            assert fail_loc["stepType"] == "failure"
            assert fail_loc["property"] == "overflow.1"

            orig_file = fail_loc["sourceLocation"]["file"]
            assert orig_file == lexer.filename
            orig_line = int(fail_loc["sourceLocation"]["line"])
            orig_cols = fail_loc["sourceLocation"]["column"].split(":")
            orig_col_start = int(orig_cols[0])
            orig_col_end   = int(orig_cols[1])

            loc = Location(
                filename  = orig_file,
                line      = orig_line,
                col_start = orig_col_start,
                col_end   = orig_col_end,
                context   = lexer.context_line[orig_line - 1])

            mh.check(loc, "operation saturates for some inputs")

            for trace in item["trace"][:-1]:
                if trace["internal"] or trace["hidden"]:
                    continue
                if trace["stepType"] == "assignment":
                    mh.info(loc,
                            "counter-example trace: %s = %s" %
                            (trace["lhs"],
                             fmt_value(trace["value"])))


class MH_BMC_Result(work_package.Result):
    def __init__(self, wp):
        super().__init__(wp, True)


class MH_BMC(command_line.MISS_HIT_Back_End):
    def __init__(self, _):
        super().__init__("MH Bounded Model Checker")

    @classmethod
    def process_wp(cls, wp):
        # Create lexer
        lexer = MATLAB_Lexer(wp.mh,
                             wp.get_content(),
                             wp.filename,
                             wp.blockname)
        if wp.cfg.octave:
            lexer.set_octave_mode()
        if not wp.cfg.pragmas:
            lexer.process_pragmas = False
        if len(lexer.text.strip()) == 0:
            return MH_BMC_Result(wp)

        # Create parse tree and translate
        try:
            parser = MATLAB_Parser(wp.mh, lexer, wp.cfg)
            n_tree = parser.parse_file()
            gst = compile_file(wp.mh, n_tree)
        except Error:
            return MH_BMC_Result(wp)

        new_filename = wp.filename.replace(".m", ".json_symtab")
        with open(new_filename, "w") as fd:
            json.dump(gst.to_json(), fd, indent=2)
        # wp.mh.info(n_tree.loc(),
        #            "wrote goto symbol table to %s" % new_filename)

        analyze(wp.mh, new_filename, lexer, n_tree)

        if not wp.options.keep:
            os.unlink(new_filename)

        return MH_BMC_Result(wp)


def main_handler():
    clp = command_line.create_basic_clp()
    clp["debug_options"].add_argument(
        "--keep",
        help="do not delete intermediate files",
        action="store_true",
        default=False)

    options = command_line.parse_args(clp)

    try:
        subprocess.run(["cbmc", "--version"],
                       check=True,
                       capture_output=True,
                       encoding="utf-8")
    except FileNotFoundError:
        clp["ap"].error("MH BMC needs 'cbmc' from the CPROVER tools on "
                        "your PATH")

    mh = Message_Handler("bmc")
    mh.show_context = not options.brief
    mh.show_style   = False
    mh.show_checks  = True
    mh.autofix      = False

    bmc_backend = MH_BMC(options)
    command_line.execute(mh, options, {}, bmc_backend)


def main():
    command_line.ice_handler(main_handler)


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2021, Florian Schanda                         ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

import datetime

from miss_hit_core import command_line
from miss_hit_core import work_package
from miss_hit_core.m_ast import *
from miss_hit_core.errors import Error, Message_Handler, ICE
from miss_hit_core.m_lexer import MATLAB_Lexer, Token_Buffer
from miss_hit_core.m_parser import MATLAB_Parser
from miss_hit_core.m_parse_utils import parse_docstrings


def get_primary_entity(options, cfg):
    if options.primary_entity:
        primary_entity = options.primary_entity
    else:
        primary_entity = cfg.style_config["copyright_primary_entity"]

    if not primary_entity and len(cfg.style_config["copyright_entity"]) == 1:
        primary_entity = list(cfg.style_config["copyright_entity"])[0]

    return primary_entity


def replace_line(lines, line_no, block_comment,
                 templates, copy_note, org, year_range,
                 offset=None):
    assert isinstance(lines, list)
    assert isinstance(line_no, int) and 1 <= line_no <= len(lines)
    assert isinstance(block_comment, bool)
    assert isinstance(templates, dict)
    assert isinstance(copy_note, str)
    assert isinstance(org, str)
    assert isinstance(year_range, tuple) and len(year_range) == 2
    assert isinstance(year_range[0], int)
    assert isinstance(year_range[1], int)
    assert offset is None or (isinstance(offset, int) and offset >= 0)

    # Replacing lines needs to do two things in order not to mess up
    # any existing formatting.
    #   1. Match the previous indentation
    #   2. Match the previous offset after the comment character

    old_line = lines[line_no - 1]
    if block_comment:
        old_indent = ""
        old_offset = re.match("^[ \t]*", old_line).group(0)
        comment_char = ""
    else:
        old_indent = re.match("^[ \t]*", old_line).group(0)
        old_offset = re.match("^[ \t]*",
                              old_line[len(old_indent) + 1:]).group(0)
        comment_char = old_line[len(old_indent)]

    if offset is not None:
        old_offset = " " * offset

    data = {"copy"   : copy_note,
            "org"    : org,
            "ystart" : year_range[0],
            "yend"   : year_range[1]}

    new_line = old_indent + comment_char + old_offset
    if year_range[0] == year_range[1]:
        new_line += templates["single"] % data
    else:
        new_line += templates["range"] % data

    lines[line_no - 1] = new_line


class MH_Copyright_Result(work_package.Result):
    def __init__(self, wp, processed=False):
        super().__init__(wp, processed)


class MH_Copyright(command_line.MISS_HIT_Back_End):
    def __init__(self):
        super().__init__("MH Copyright")

    @classmethod
    def process_wp(cls, wp):
        # Load file content
        content = wp.get_content()

        # Create lexer
        lexer = MATLAB_Lexer(wp.mh, content, wp.filename, wp.blockname)
        if wp.cfg.octave:
            lexer.set_octave_mode()
        if not wp.cfg.pragmas:  # pragma: no cover
            lexer.process_pragmas = False

        # Deal with command-line options, setting up three variables:
        # * action - what to do
        # * primary_entity - main entity
        # * templates - how to write copyright notices
        templates = {"single" : wp.options.template,
                     "range"  : wp.options.template_range}
        primary_entity = get_primary_entity(wp.options, wp.cfg)
        if not primary_entity:
            wp.mh.warning(lexer.get_file_loc(),
                          "unable to determine primary copyright entity,"
                          " skipping this file")
            return MH_Copyright_Result(wp, False)
        if wp.options.update_year:
            action = "update_year"
        elif wp.options.merge:
            action = "merge"
        elif wp.options.change_entity is not None:
            action = "change_entity"
        elif wp.options.add_notice:
            action = "add_notice"
        else:
            raise ICE("none of the action are set")

        allowed_entities = (wp.cfg.style_config["copyright_entity"] |
                            set([primary_entity]))

        # We're dealing with an empty file here. Lets just not do anything
        if len(lexer.text.strip()) == 0:
            return MH_Copyright_Result(wp, False)

        # Parse
        try:
            tbuf = Token_Buffer(lexer, wp.cfg)
            parser = MATLAB_Parser(wp.mh, tbuf, wp.cfg)
            parse_tree = parser.parse_file()
            parse_docstrings(wp.mh, wp.cfg, parse_tree, tbuf)
        except Error:  # pragma: no cover
            return MH_Copyright_Result(wp, False)

        # Determine the docstring of the primary entity. We use the
        # script/class docstring, if it exists and contains copyright
        # info. Otherwise we use the compilation unit's docstring
        # (i.e. file header). The style checker worries about the case
        # where we have copyright in more than one location.
        n_docstring = None
        if isinstance(parse_tree, Class_File):
            n_docstring = parse_tree.n_classdef.n_docstring
        elif isinstance(parse_tree, Function_File):
            n_docstring = parse_tree.l_functions[0].n_docstring
        if n_docstring is None or not n_docstring.copyright_info:
            n_docstring = parse_tree.n_docstring
        # Note that n_docstring could be None at this point, so we do
        # need to deal with that correctly.

        # Set up line -> Copyright_Info map
        cinfos = {}
        if n_docstring:
            for cinfo in n_docstring.copyright_info:
                if cinfo.t_comment.location.line in cinfos:
                    raise ICE("docstring with duplicate line (%u)" %
                              cinfo.t_comment.location.line)
                else:
                    cinfos[cinfo.t_comment.location.line] = cinfo

            # The merge action requires the copyright notice to be
            # grouped. Enforce this now.
            if action == "merge" and \
               not (n_docstring.all_copyright_in_one_block(allowed_entities) or
                    n_docstring.all_copyright_in_one_block()):
                wp.mh.error(n_docstring.loc(),
                            "cannot merge entries in this docstring as they"
                            " are not all next to each other",
                            fatal=False)
                return MH_Copyright_Result(wp, False)

        # Perform actions
        action_taken = False
        lines = copy(lexer.context_line)
        try:
            if action == "update_year":
                new_year = wp.options.year
                for line_no, cinfo in cinfos.items():
                    ystart, yend = cinfo.get_range()
                    if cinfo.get_org() != primary_entity:
                        continue

                    if yend <= new_year:
                        new_range = (ystart, new_year)
                        replace_line(lines, line_no, cinfo.is_block_comment(),
                                     templates, cinfo.get_copy_notice(),
                                     primary_entity, new_range)

                        action_taken = True
                    else:
                        wp.mh.error(cinfo.loc_yend(),
                                    "end year is later than %s" % new_year)

            elif action == "merge":
                merged_line = None
                merged_line_is_block = None
                killed_lines = []
                new_range = None
                for line_no in sorted(cinfos):
                    cinfo = cinfos[line_no]
                    ystart, yend = cinfo.get_range()

                    if cinfo.get_org() not in allowed_entities:
                        continue
                    if cinfo.get_org() != primary_entity or ystart == yend:
                        action_taken = True

                    if ystart > yend:
                        wp.mh.error(cinfo.loc_ystart(),
                                    "initial year is later than end year")

                    if new_range is None:
                        merged_line = line_no
                        merged_line_is_block = cinfo.is_block_comment()
                        merged_line_copy_notice = cinfo.get_copy_notice()
                        new_range = cinfo.get_range()
                    else:
                        new_range = (min(ystart, new_range[0]),
                                     max(yend, new_range[1]))
                        killed_lines.append(line_no)
                        action_taken = True

                if action_taken:
                    replace_line(lines, merged_line, merged_line_is_block,
                                 templates, merged_line_copy_notice,
                                 primary_entity, new_range)
                    for line_no in reversed(killed_lines):
                        del lines[line_no - 1]

            elif action == "change_entity":
                for line_no, cinfo in cinfos.items():
                    if cinfo.get_org() == wp.options.change_entity:
                        replace_line(lines, line_no, cinfo.is_block_comment(),
                                     templates, cinfo.get_copy_notice(),
                                     primary_entity,
                                     cinfo.get_range())
                        action_taken = True

            elif action == "add_notice":
                # Only add copyright notices to files without any
                # notice.
                if len(cinfos) == 0:
                    # Find the right place to add the notice. Right
                    # now we only do this for the file header; so we
                    # need to find it and append at the end; or create
                    # a new one.
                    action_taken = True

                    if parse_tree.n_docstring:
                        is_block, line_no = parse_tree.n_docstring.final_line()
                        off = parse_tree.n_docstring.guess_docstring_offset()

                        # Duplicate the last line
                        lines = \
                            lines[:line_no] + \
                            [lines[line_no - 1]] + \
                            lines[line_no:]
                    else:
                        is_block = False
                        line_no = 0
                        off = 1
                        lines = ["% POTATO", ""] + lines

                    if wp.options.style == "c_first" or \
                       (wp.options.style == "dynamic" and
                        not wp.cfg.octave):
                        copy_notice = "(c) Copyright"
                    else:
                        copy_notice = "Copyright (c)"

                    # Then overwrite it
                    replace_line(lines, line_no + 1, is_block,
                                 templates, copy_notice, primary_entity,
                                 (wp.options.year, wp.options.year),
                                 offset = off)

            else:
                raise ICE("Unexpected action '%s'" % action)
        except Error:
            return MH_Copyright_Result(wp, False)

        if action_taken:
            wp.write_modified("\n".join(lines) + "\n")

        return MH_Copyright_Result(wp, True)


def main_handler():
    clp = command_line.create_basic_clp(
        epilog=("Remember to carefully review any code changes this tool"
                " makes, as copyright notices are generally considered to"
                " be quite important."))

    clp["ap"].add_argument("--process-slx",
                           action="store_true",
                           default=False,
                           help=("Update copyright notices inside Simulink"
                                 " models. This option is temporary, and"
                                 " will be removed in future once the"
                                 " feature is good enough to be enabled"
                                 " by default."))

    c_actions = clp["ap"].add_argument_group("copyright action")
    c_actions = c_actions.add_mutually_exclusive_group(required=True)
    c_actions.add_argument("--update-year",
                           action="store_true",
                           default=False,
                           help=("Update the end year in copyright notices for"
                                 " the primary copyright holder to the current"
                                 " year."))
    c_actions.add_argument("--merge",
                           action="store_true",
                           default=False,
                           help=("Merge all non-3rd party copyright notices"
                                 " into one for the primary copyright holder"))
    c_actions.add_argument("--change-entity",
                           default=None,
                           metavar="OLD_COPYRIGHT_HOLDER",
                           help=("Change notices from the specified copyright"
                                 " holder into the primary copyright holder."))
    c_actions.add_argument("--add-notice",
                           action="store_true",
                           default=False,
                           help=("Add a copyright notice to files that do not"
                                 " have one yet."))
    c_actions.add_argument("--style",
                           choices=("dynamic", "c_first", "c_last"),
                           default="dynamic",
                           help=("'(c) Copyright' or 'Copyright (c)'. Dynamic"
                                 " picks the existing style, and if there is"
                                 " nothing, picks c_first for MATLAB code and"
                                 " c_last for Octave code."
                                 " Default is %(default)s."))

    c_data = clp["ap"].add_argument_group("copyright data")
    c_data.add_argument("--year",
                        default=datetime.datetime.now().year,
                        type=int,
                        help=("The current year (by default this is"
                              " %(default)s)"))
    c_data.add_argument("--primary-entity",
                        default=None,
                        metavar="COPYRIGHT_HOLDER",
                        help=("The primary copyright entity."))
    c_data.add_argument("--template-range",
                        default="%(copy)s %(ystart)u-%(yend)u %(org)s",
                        metavar="TEMPLATE_TEXT",
                        help=("Text template to use for a copyright notice"
                              " with a year range. default: '%(default)s'"))
    c_data.add_argument("--template",
                        default="%(copy)s %(yend)u %(org)s",
                        metavar="TEMPLATE_TEXT",
                        help=("Text template to use for a copyright notice"
                              " with a single year. default: '%(default)s'"))

    options = command_line.parse_args(clp)

    # Sanity check year
    if options.year < 1000:  # pragma: no cover
        clp["ap"].error("year must be at lest 1000")
    elif options.year >= 10000:  # pragma: no cover
        clp["ap"].error("I am extremely concerned that this tool is still"
                        " useful after 8000 years, stop what you're doing"
                        " and change programming language NOW.")

    mh = Message_Handler("copyright")

    mh.show_context = not options.brief
    mh.show_style   = False
    mh.show_checks  = False
    mh.autofix      = True

    copyright_backend = MH_Copyright()
    command_line.execute(mh, options, {},
                         copyright_backend,
                         options.process_slx)


def main():
    command_line.ice_handler(main_handler)


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2020-2021, Florian Schanda                    ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify                    ##
##  it under the terms of the GNU Affero General Public License as          ##
##  published by the Free Software Foundation, either version 3 of the      ##
##  License, or (at your option) any later version.                         ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU Afferto General Public License for more details.                    ##
##                                                                          ##
##  You should have received a copy of the GNU Affero General Public        ##
##  License along with MISS_HIT. If not, see                                ##
##  <http://www.gnu.org/licenses/>.                                         ##
##                                                                          ##
##############################################################################

import os

from miss_hit_core.m_ast import *
from miss_hit_core.errors import Message_Handler, ICE
from miss_hit_core.cfg_ast import Library_Declaration, Entrypoint_Declaration
from miss_hit_core import cfg_tree

from miss_hit.m_entity import *


def treewalk(n_root, function):
    assert isinstance(n_root, Node)

    class Visitor(AST_Visitor):
        def visit(self, node, n_parent, relation):
            function(node)

    n_root.visit(None, Visitor(), "Root")


class Semantic_Analysis_Pass_1:
    def __init__(self, mh):
        assert isinstance(mh, Message_Handler)
        self.mh    = mh
        self.scope = Scope()
        self.pkg   = []

    def sem_compilation_unit(self, n_cu):
        assert isinstance(n_cu, Compilation_Unit)

        n_cu.scope = self.scope

        if isinstance(n_cu, Class_File):
            self.sem_class_file(n_cu)
        elif isinstance(n_cu, Function_File):
            pass
        elif isinstance(n_cu, Script_File):
            pass
        else:
            raise ICE("unexpected compilation unit type %s" %
                      n_cu.__class__.__name__)

    def sem_class_file(self, n_cf):
        assert isinstance(n_cf, Class_File)

        e_cls = Class_Entity(n_cf.n_classdef)
        self.scope.register(self.mh, e_cls)


def sem_pass_1(mh, entrypoint, n_cu):
    assert isinstance(mh, Message_Handler)
    assert isinstance(n_cu, Compilation_Unit)
    assert isinstance(entrypoint, (Library_Declaration,
                                   Entrypoint_Declaration)) or \
        entrypoint is None

    rv = Semantic_Analysis_Pass_1(mh)
    rv.sem_compilation_unit(n_cu)

    item = os.path.normpath(n_cu.dirname)
    if entrypoint:
        best_match = None
        for path in cfg_tree.get_source_path(entrypoint):
            search_item = os.path.normpath(path)
            if item.startswith(search_item):
                if best_match is None or len(best_match) < len(search_item):
                    best_match = search_item
        if best_match is None:
            raise ICE("could not find %s on path" % n_cu.dirname)
        packages = item[len(best_match) + 1:]
        if os.sep != "/":
            packages = packages.replace(os.sep, "/")
        if packages:
            packages = packages.split("/")
            rv.pkg = packages
        else:
            packages = []

        # We have an empty list, or a list of + directories, followed
        # by at most one @ directory, followed optionally by a private
        # directory.
        current_dir = best_match
        sequence = "package"

        for item in packages:
            current_dir = os.path.join(current_dir, item)
            if item.startswith("+"):
                if sequence == "package":
                    pass
                else:
                    mh.check(n_cu.loc(),
                             "cannot nest package inside a %s" % sequence)
                    return None

            elif item.startswith("@"):
                if sequence == "package":
                    sequence = "class directory"

                else:
                    mh.check(n_cu.loc(),
                             "cannot nest class directory inside a %s" %
                             sequence)
                    return None

            elif item == "private":
                if sequence in ("package", "class directory"):
                    sequence = "private directory"

                else:
                    mh.check(n_cu.loc(),
                             "cannot private directory inside a %s" %
                             sequence)
                    return None

            else:
                mh.check(n_cu.loc(),
                         "is not on path and cannot be accessed" %
                         sequence)
                return None

    return rv
